/*
 * ProActive Parallel Suite(TM):
 * The Open Source library for parallel and distributed
 * Workflows & Scheduling, Orchestration, Cloud Automation
 * and Big Data Analysis on Enterprise Grids & Clouds.
 *
 * Copyright (c) 2007 - 2017 ActiveEon
 * Contact: contact@activeeon.com
 *
 * This library is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License
 * as published by the Free Software Foundation: version 3 of
 * the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * If needed, contact us to obtain a release under GPL Version 2 or 3
 * or a different license than the AGPL.
 */
package org.ow2.proactive.topology.descriptor;

import java.io.Serializable;

import org.objectweb.proactive.annotation.PublicAPI;


/**
 *
 * Class represents the descriptor of the nodes topology (network location)
 * which could be used with {@code ResourceManager.getAtMostNodes} request.
 * <p>
 * User may create its own instance of available descriptors and parameterize it or
 * use one of predefined constants when it is sufficient.
 */
@PublicAPI
public class TopologyDescriptor implements Serializable {

    /** no constraint on node location */
    public static final TopologyDescriptor ARBITRARY = new ArbitraryTopologyDescriptor();

    /** the set of closest nodes */
    public static final TopologyDescriptor BEST_PROXIMITY = new BestProximityDescriptor();

    /** the set of nodes on a single host */
    public static final TopologyDescriptor SINGLE_HOST = new SingleHostDescriptor();

    /**
     * the set of nodes of a single host exclusively
     * (the host is reserved for the user)
     */
    public static final TopologyDescriptor SINGLE_HOST_EXCLUSIVE = new SingleHostExclusiveDescriptor();

    /**
     * the set of nodes of several hosts exclusively
     * (hosts are reserved for the user)
     */
    public static final TopologyDescriptor MULTIPLE_HOSTS_EXCLUSIVE = new MultipleHostsExclusiveDescriptor();

    /**
     * the set of nodes one per host exclusively
     * (hosts are reserved for the user)
     */
    public static final TopologyDescriptor DIFFERENT_HOSTS_EXCLUSIVE = new DifferentHostsExclusiveDescriptor();

    /**
     * the flag indicated that descriptor requires the topology information in the resource manager.
     * It affects the scripts execution strategy: if true selection scripts are executed on all nodes
     * first and then the topology information is taken into account. If false scripts are executed
     * only on required number of nodes.
     * If this field is set to false the descriptor could be used even when the topology is
     * disabled in the resource manager.
     */
    private boolean topologyBased;

    /**
     * Creates the descriptor.
     * @param topologyBased indicates that descriptor requires the topology information in the resource manager.
     * It affects the scripts execution strategy: if true selection scripts are executed on all nodes
     * first and then the topology information is taken into account. If false scripts are executed
     * only on required number of nodes.
     * If this field is set to false the descriptor could be used even when the topology is
     * disabled in the resource manager.
     */
    protected TopologyDescriptor(boolean topologyBased) {
        this.topologyBased = topologyBased;
    }

    /**
     * Defines if selection scripts have to be executed on all
     * available nodes before the topology information will be processed.
     *
     * @return true in case of greedy behavior false otherwise
     */
    public boolean isTopologyBased() {
        return topologyBased;
    }

    /**
     * Returns the string representation of the descriptor.
     */
    public String toString() {
        return this.getClass().getSimpleName();
    }
}
