/*
 * ProActive Parallel Suite(TM):
 * The Open Source library for parallel and distributed
 * Workflows & Scheduling, Orchestration, Cloud Automation
 * and Big Data Analysis on Enterprise Grids & Clouds.
 *
 * Copyright (c) 2007 - 2017 ActiveEon
 * Contact: contact@activeeon.com
 *
 * This library is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License
 * as published by the Free Software Foundation: version 3 of
 * the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * If needed, contact us to obtain a release under GPL Version 2 or 3
 * or a different license than the AGPL.
 */
package org.ow2.proactive.resourcemanager.utils;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.nio.charset.StandardCharsets;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import javax.management.InstanceNotFoundException;
import javax.management.IntrospectionException;
import javax.management.MBeanAttributeInfo;
import javax.management.MBeanServer;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.management.ReflectionException;
import javax.management.openmbean.CompositeDataSupport;

import org.apache.log4j.Logger;
import org.objectweb.proactive.utils.StackTraceUtil;
import org.ow2.proactive.jmx.RRDDataStore;
import org.rrd4j.core.RrdDb;
import org.rrd4j.core.Sample;


/**
 * This class (thread) dump all properties of Sigar MBean to RRD data base with specific period.
 */
public class RRDSigarDataStore extends RRDDataStore {

    private static final String[] OBJECT_NAMES = { "java.lang:type=OperatingSystem", "java.lang:type=Memory",
                                                   "java.lang:type=Threading", "java.lang:type=ClassLoading",
                                                   "sigar:Type=CpuCoreUsage,Name=*", "sigar:Type=FileSystem,Name=*",
                                                   "sigar:Type=CpuUsage", "sigar:Type=Mem",
                                                   "sigar:Type=NetInterface,Name=*", "sigar:Type=Swap" };

    private HashMap<String, String> compositeTypes = new HashMap<>();

    private static Set<String> exceptionsThrown = Collections.synchronizedSet(new HashSet<String>());

    private MBeanServer mbs;

    /**
     * Initializes a new RRD data base if it's not exist.
     *
     * @param mbs            is the source of chronological data
     * @param dataBaseFilePath is the path to the file with the rrd data base
     * @param step             is the data base refresh period
     * @throws java.io.IOException is thrown when the data base exists but cannot be read
     */
    public RRDSigarDataStore(MBeanServer mbs, String dataBaseFilePath, int step, Logger logger) throws IOException,
            MalformedObjectNameException, IntrospectionException, InstanceNotFoundException, ReflectionException {
        super(dataBaseFilePath, step, logger);

        compositeTypes.put("HeapMemoryUsage-java.lang:type=Memory", "used");

        this.mbs = mbs;

        for (String objectName : OBJECT_NAMES) {
            Set<ObjectName> realNames = mbs.queryNames(new ObjectName(objectName), null);

            for (ObjectName realObjectName : realNames) {
                for (MBeanAttributeInfo attrInfo : mbs.getMBeanInfo(realObjectName).getAttributes()) {
                    if (attrInfo.isReadable()) {

                        String sourceName = attrInfo.getName() + "-" + realObjectName;
                        String dataStoreName = toDataStoreName(sourceName);

                        logger.trace("Adding a " + dataStoreName + " / " + sourceName + " source to rrd");
                        dataSources.put(dataStoreName, sourceName);
                    }
                }
            }
        }

        initDatabase();
        setName("RRD4J Data Store " + new File(dataBaseFile).getName());

        setDaemon(true);
        start();
    }

    /**
     * Shorten the name to 20 symbols trying to keep it meaningful
     */
    public static String toDataStoreName(String name) {
        String res = name.replace("-java.lang:type=", "");

        // to unify the following
        // Combined-sigar:Name=4,Type=CpuCoreUsage
        // Combined-sigar:Type=CpuCoreUsage,Name=4
        res = res.replace(",Type=CpuCoreUsage", "");
        res = res.replace("Type=CpuCoreUsage,", "");

        // RxBytes-sigar:Name=lo,Type=NetInterface / RxBytesName=lo,Type
        // RxBytes-sigar:Type=NetInterface,Name=eth0
        res = res.replace(",Type=NetInterface", "");
        res = res.replace("Type=NetInterface,", "");

        res = res.replace("-sigar:Type=", "");
        res = res.replace("-sigar:", "");

        if (res.length() > 20) {
            // only 20 symbols allowed in data source name
            res = res.substring(0, 19);
        }
        return res;
    }

    /**
     * Periodically dumps the new mbean state to the data base
     */
    public void run() {
        try {

            RrdDb dataBase = new RrdDb(dataBaseFile);

            logger.debug("RRD database configuration:\n" + dataBase.getRrdDef().dump());

            while (!terminate) {
                synchronized (dataSources) {

                    dataSources.wait(step * 1000);

                    if (terminate) {
                        break;
                    }
                    sample(dataBase, System.currentTimeMillis());
                }
            }
            dataBase.close();
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
        }
    }

    // package protected for testing
    void sample(RrdDb dataBase, long timeInMs) throws IOException {
        Sample sample = dataBase.createSample();

        // updating the data base
        for (String dataSource : dataSources.keySet()) {
            String fullName = dataSources.get(dataSource);
            String[] names = fullName.split("-");
            String attrName = names[0];
            String objectName = names[1];

            // Redirect error stream to avoid sigar exceptions flooding
            // @see https://github.com/ow2-proactive/scheduling/issues/3068
            StdErrRedirectionHelper stdErrRedirectionHelper = new StdErrRedirectionHelper();

            stdErrRedirectionHelper.redirectStdErr();

            readAttribute(timeInMs, sample, dataSource, fullName, attrName, objectName);

            String errorContent = stdErrRedirectionHelper.readAndReset();

            if (errorContent.length() > 0) {
                logErrorOnce(errorContent,
                             "Error while reading attribute " + attrName + " for object " + objectName + ": " +
                                           System.lineSeparator() + errorContent);
            }
        }

        try {
            sample.setTime(timeInMs / 1000);
            sample.update();
        } catch (Exception e) {
            logger.error("Cannot update RRD database: " + e.getMessage());
        }
    }

    private void readAttribute(long timeInMs, Sample sample, String dataSource, String fullName, String attrName,
            String objectName) {
        try {

            Object attrValue = mbs.getAttribute(new ObjectName(objectName), attrName);
            if (attrValue instanceof CompositeDataSupport && compositeTypes.get(fullName) != null) {
                Object val = ((CompositeDataSupport) attrValue).get(compositeTypes.get(fullName));
                if (val != null) {
                    attrValue = val;
                }
            }
            sample.setValue(dataSource, Double.parseDouble(attrValue.toString()));
            if (logger.isTraceEnabled()) {
                logger.trace(timeInMs / 1000 + " sampling: " + dataSource + " / " + fullName + " " +
                             Double.parseDouble(attrValue.toString()));
            }
        } catch (NumberFormatException ex) {
            // do not save non-numeric values
            String exceptionAsString = StackTraceUtil.getStackTrace(ex).replaceAll("For input string: \".*\"", "");
            exceptionAsString = fullName + " / " + exceptionAsString;

            logErrorOnce(exceptionAsString,
                         "Non numeric value for " + dataSource + " / " + fullName + ": " + ex.getMessage());
        } catch (Exception e) {
            String exceptionAsString = StackTraceUtil.getStackTrace(e);
            logErrorOnce(exceptionAsString,
                         "Error while reading attribute " + attrName + " for object " + objectName + ": ",
                         e);
        }
    }

    /**
     * Logs a sigar error only once to avoid flooding
     * @param exceptionAsString key used to memorize exceptions already handled
     * @param message message to display
     */
    private void logErrorOnce(String exceptionAsString, String message) {
        if (!exceptionsThrown.contains(exceptionAsString)) {
            exceptionsThrown.add(exceptionAsString);
            logger.warn(message);
        } else if (logger.isTraceEnabled()) {
            logger.trace(message);
        }
    }

    /**
     * Logs a sigar error only once to avoid flooding
     * @param exceptionAsString key used to memorize exceptions already handled
     * @param message message to display
     * @param e real exception
     */
    private void logErrorOnce(String exceptionAsString, String message, Exception e) {
        if (!exceptionsThrown.contains(exceptionAsString)) {
            exceptionsThrown.add(exceptionAsString);
            logger.warn(message, e);
        } else if (logger.isTraceEnabled()) {
            logger.trace(message, e);
        }
    }

    private class StdErrRedirectionHelper {
        private ByteArrayOutputStream storedErrors;

        private PrintStream redirectedStream;

        private PrintStream console;

        public PrintStream redirectStdErr() throws UnsupportedEncodingException {
            storedErrors = new ByteArrayOutputStream();
            redirectedStream = new PrintStream(storedErrors, true, StandardCharsets.UTF_8.name());
            console = System.err;
            System.setErr(redirectedStream);
            return redirectedStream;
        }

        public String readAndReset() {
            System.setErr(console);
            String errorContent = new String(storedErrors.toByteArray(), StandardCharsets.UTF_8);
            redirectedStream.close();
            return errorContent.trim();
        }
    }
}
