/*
 * ProActive Parallel Suite(TM):
 * The Open Source library for parallel and distributed
 * Workflows & Scheduling, Orchestration, Cloud Automation
 * and Big Data Analysis on Enterprise Grids & Clouds.
 *
 * Copyright (c) 2007 - 2017 ActiveEon
 * Contact: contact@activeeon.com
 *
 * This library is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License
 * as published by the Free Software Foundation: version 3 of
 * the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * If needed, contact us to obtain a release under GPL Version 2 or 3
 * or a different license than the AGPL.
 */
package org.ow2.proactive.scheduler.task;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.Serializable;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Pattern;

import org.apache.log4j.*;
import org.hamcrest.Matchers;
import org.junit.After;
import org.junit.Assert;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;
import org.ow2.proactive.scheduler.common.job.JobId;
import org.ow2.proactive.scheduler.common.task.Log4JTaskLogs;
import org.ow2.proactive.scheduler.common.task.TaskId;
import org.ow2.proactive.scheduler.common.util.TaskLoggerRelativePathGenerator;
import org.ow2.proactive.scheduler.common.util.logforwarder.AppenderProvider;
import org.ow2.proactive.scheduler.common.util.logforwarder.LogForwardingException;
import org.ow2.proactive.scheduler.job.JobIdImpl;
import org.ow2.proactive.scripting.ScriptResult;
import org.ow2.proactive.scripting.SimpleScript;
import org.ow2.proactive.scripting.TaskScript;
import org.ow2.tests.ProActiveTestClean;


public class TaskLoggerTest extends ProActiveTestClean {

    @Rule
    public TemporaryFolder tempFolder = new TemporaryFolder();

    private TaskLogger taskLogger;

    @After
    public void removeLogger() {
        if (taskLogger != null) {
            taskLogger.close();
            assertNull(LogManager.exists(taskLogger.getName()));
        }
    }

    @Test
    public void printAndGetLogs() throws Exception {

        TaskLogger taskLogger = new TaskLogger(TaskIdImpl.createTaskId(new JobIdImpl(1000, "job"), "task", 42L),
                                               "myhost");

        assertEquals("", taskLogger.getLogs().getAllLogs(false));

        taskLogger.getOutputSink().println("hello");

        System.out.println(taskLogger.getLogs().getAllLogs(false));

        assertEquals(String.format("hello%n"), taskLogger.getLogs().getAllLogs(false));
        assertEquals(String.format("hello%n"), taskLogger.getLogs().getStdoutLogs(false));

        taskLogger.getErrorSink().println("error");
        assertEquals(String.format("hello%nerror%n"), taskLogger.getLogs().getAllLogs(false));
        assertEquals(String.format("error%n"), taskLogger.getLogs().getStderrLogs(false));
    }

    @Test
    public void logStreaming() throws Exception {
        taskLogger = new TaskLogger(TaskIdImpl.createTaskId(new JobIdImpl(1000, "job"), "task", 42L), "myhost");

        final StringWriter stringAppender = new StringWriter();
        AppenderProvider stringAppenderProvider = new AppenderProvider() {
            @Override
            public Appender getAppender() throws LogForwardingException {
                return new WriterAppender(new PatternLayout("%m%n"), stringAppender);
            }
        };

        taskLogger.activateLogs(stringAppenderProvider);

        assertEquals("", stringAppender.toString());

        taskLogger.getOutputSink().println("hello");

        // async appender, it gets buffered
        assertEquals("", stringAppender.toString());

        for (int i = 0; i < 1000; i++) {
            taskLogger.getOutputSink().println("hello");
        }
        assertNotEquals("", stringAppender.toString());
    }

    @Test
    public void getStoredLogs() throws Exception {
        taskLogger = new TaskLogger(TaskIdImpl.createTaskId(new JobIdImpl(1000, "job"), "task", 42L), "myhost");

        final StringWriter stringAppender = new StringWriter();
        AppenderProvider stringAppenderProvider = new AppenderProvider() {
            @Override
            public Appender getAppender() throws LogForwardingException {
                return new WriterAppender(new PatternLayout("%m%n"), stringAppender);
            }
        };

        taskLogger.getStoredLogs(stringAppenderProvider);
        assertEquals("", stringAppender.toString());

        taskLogger.getOutputSink().println("hello");

        taskLogger.getStoredLogs(stringAppenderProvider);
        assertEquals(String.format("hello%n%n"), stringAppender.toString());
    }

    @Test
    public void logPattern() throws Exception {
        TaskId taskId = TaskIdImpl.createTaskId(new JobIdImpl(1000, "job"), "task", 42L);
        taskLogger = new TaskLogger(taskId, "myhost");

        assertEquals("", taskLogger.getLogs().getAllLogs(false));

        taskLogger.getOutputSink().println("hello");

        String quotedStringTaskId = Pattern.quote(taskId.toString());
        assertTrue(taskLogger.getLogs()
                             .getAllLogs(true)
                             .matches("\\[" + quotedStringTaskId +
                                      "@myhost;[0-9][0-9]:[0-9][0-9]:[0-9][0-9]\\] hello\r?\n"));

        taskLogger.getErrorSink().println("error");
        assertTrue(taskLogger.getLogs()
                             .getStderrLogs(true)
                             .matches("\\[" + quotedStringTaskId +
                                      "@myhost;[0-9][0-9]:[0-9][0-9]:[0-9][0-9]\\] error\r?\n"));

    }

    @Test
    public void createFileAppenderTest() throws Exception {
        JobId jobid = new JobIdImpl(1000, "job");
        TaskId taskId = TaskIdImpl.createTaskId(jobid, "task", 42L);
        taskLogger = new TaskLogger(taskId, "myhost");

        File logFolder = tempFolder.newFolder("logfolder");

        File logAppender = taskLogger.createFileAppender(logFolder);

        System.out.println(logAppender.getParentFile().getName());

        assertEquals(logAppender.getParentFile().getName(), jobid.value());

        assertEquals(logAppender.getName(), new TaskLoggerRelativePathGenerator(taskId).getFileName());

    }

    @Test
    public void multipleRemoveLoggers() throws Exception {
        List<TaskLogger> loggers = new ArrayList<>(1000);
        for (int i = 0; i < 1000; i++) {
            loggers.add(new TaskLogger(TaskIdImpl.createTaskId(new JobIdImpl(1000, "job"), "task", i), "myhost"));
        }

        for (int i = 0; i < 1000; i++) {
            loggers.get(i).close();
            assertNull(LogManager.exists(loggers.get(i).getName()));
        }
        assertNull(LogManager.exists(Log4JTaskLogs.getLoggerName("" + 1000)));
    }

    @Test
    public void testScalaOutputFromScript() throws Exception {
        BasicConfigurator.resetConfiguration();
        BasicConfigurator.configure();
        Logger.getRootLogger().setLevel(Level.INFO);
        String scalaScript = "println(\"Hello World\")" + System.getProperty("line.separator");
        scalaScript += "print(\"How are you\")";
        TaskLogger taskLogger = new TaskLogger(TaskIdImpl.createTaskId(new JobIdImpl(1, "job1"), "task1", 1), "myhost");
        SimpleScript ss = new SimpleScript(scalaScript, "scalaw");
        TaskScript taskScript = new TaskScript(ss);
        ScriptResult<Serializable> res = taskScript.execute(null,
                                                            taskLogger.getOutputSink(),
                                                            taskLogger.getErrorSink());
        taskLogger.close();
        Assert.assertThat(res.getOutput(), Matchers.containsString("Hello World"));
        Assert.assertThat(taskLogger.getLogs().getAllLogs(), Matchers.containsString("Hello World"));
        Assert.assertThat(res.getOutput(), Matchers.containsString("How are you"));
        Assert.assertThat(taskLogger.getLogs().getAllLogs(), Matchers.containsString("How are you"));

    }
}
