/*
 * ProActive Parallel Suite(TM):
 * The Open Source library for parallel and distributed
 * Workflows & Scheduling, Orchestration, Cloud Automation
 * and Big Data Analysis on Enterprise Grids & Clouds.
 *
 * Copyright (c) 2007 - 2017 ActiveEon
 * Contact: contact@activeeon.com
 *
 * This library is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License
 * as published by the Free Software Foundation: version 3 of
 * the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * If needed, contact us to obtain a release under GPL Version 2 or 3
 * or a different license than the AGPL.
 */
package functionaltests.service;

import static org.junit.Assert.assertEquals;

import java.util.Map;

import org.junit.Assert;
import org.junit.Test;
import org.ow2.proactive.scheduler.common.JobDescriptor;
import org.ow2.proactive.scheduler.common.SchedulerEvent;
import org.ow2.proactive.scheduler.common.TaskDescriptor;
import org.ow2.proactive.scheduler.common.job.JobId;
import org.ow2.proactive.scheduler.common.job.TaskFlowJob;
import org.ow2.proactive.scheduler.common.task.JavaTask;
import org.ow2.proactive.scheduler.common.task.NativeTask;
import org.ow2.proactive.scheduler.common.task.OnTaskError;
import org.ow2.proactive.scheduler.descriptor.EligibleTaskDescriptor;
import org.ow2.proactive.scheduler.descriptor.JobDescriptorImpl;
import org.ow2.proactive.scheduler.job.InternalJob;
import org.ow2.proactive.scheduler.task.TaskResultImpl;
import org.ow2.proactive.scheduler.task.internal.InternalTask;


public class SchedulingServiceTest2 extends BaseServiceTest {

    private TaskFlowJob createTestJob() throws Exception {
        TaskFlowJob job = new TaskFlowJob();
        job.setName(this.getClass().getSimpleName());

        JavaTask task1 = new JavaTask();
        task1.setName("javaTask");
        task1.setExecutableClassName("class");
        task1.setOnTaskError(OnTaskError.CANCEL_JOB);
        job.addTask(task1);

        NativeTask task2 = new NativeTask();
        task2.setName("nativeTask");
        task2.setCommandLine("command line");
        task2.setOnTaskError(OnTaskError.CANCEL_JOB);
        job.addTask(task2);

        return job;
    }

    @Test
    public void testFailedNativeTask() throws Exception {
        testFailedTask(false);
    }

    @Test
    public void testFailedJavaTask() throws Exception {
        testFailedTask(true);
    }

    @Test
    public void testNotStartedJob() throws Exception {
        InternalJob job = createJob(createTestJob());
        service.submitJob(job);
        listener.assertEvents(SchedulerEvent.JOB_SUBMITTED);

        service.killJob(job.getId());
        listener.assertEvents(SchedulerEvent.TASK_RUNNING_TO_FINISHED,
                              SchedulerEvent.TASK_RUNNING_TO_FINISHED,
                              SchedulerEvent.JOB_PENDING_TO_FINISHED,
                              SchedulerEvent.JOB_UPDATED);
    }

    private void testFailedTask(boolean failNativeTask) throws Exception {
        service.submitJob(createJob(createTestJob()));
        listener.assertEvents(SchedulerEvent.JOB_SUBMITTED);

        Map<JobId, JobDescriptor> jobsMap;
        JobDescriptor jobDesc;

        jobsMap = service.lockJobsToSchedule();

        assertEquals(1, jobsMap.size());
        jobDesc = jobsMap.values().iterator().next();
        Assert.assertEquals(2, jobDesc.getEligibleTasks().size());

        for (TaskDescriptor taskDesc : jobDesc.getEligibleTasks()) {
            taskStarted(jobDesc, (EligibleTaskDescriptor) taskDesc);
        }
        service.unlockJobsToSchedule(jobsMap.values());

        if (failNativeTask) {
            InternalTask nativeTask = ((JobDescriptorImpl) jobDesc).getInternal().getTask("nativeTask");
            // native task terminates with code 1, flag 'cancelJobOnError' was
            // set so job should be cancelled
            service.taskTerminatedWithResult(nativeTask.getId(),
                                             new TaskResultImpl(nativeTask.getId(), new RuntimeException(), null, 0));
        } else {
            InternalTask javaTask = ((JobDescriptorImpl) jobDesc).getInternal().getTask("javaTask");
            // java task terminates with exception, flag 'cancelJobOnError' was
            // set so job should be cancelled
            service.taskTerminatedWithResult(javaTask.getId(),
                                             new TaskResultImpl(javaTask.getId(), new RuntimeException(), null, 0));
        }

        jobsMap = service.lockJobsToSchedule();
        assertEquals(0, jobsMap.size());

        listener.assertEvents(SchedulerEvent.JOB_PENDING_TO_RUNNING,
                              SchedulerEvent.JOB_UPDATED,
                              SchedulerEvent.TASK_PENDING_TO_RUNNING,
                              SchedulerEvent.TASK_PENDING_TO_RUNNING,
                              SchedulerEvent.TASK_RUNNING_TO_FINISHED,
                              SchedulerEvent.TASK_RUNNING_TO_FINISHED,
                              SchedulerEvent.JOB_RUNNING_TO_FINISHED,
                              SchedulerEvent.JOB_UPDATED);

        infrastructure.assertRequests(2);
    }

}
